# Chemical reactions and chemical species

This language extension implements the domain concept of chemical reactions. The language enables defining chemical species and reactions, and computing thermochemical and electrochemical properties for the defined species and reactions.

## Chemical species

The syntax for a species definition is:

```
Species <name> [table with parameters and properties]
```

The word `Species` is a keyword, the species `name` is mandatory and the table is optional. Here an example for water:

```
water = Species H2O
```

Currently, the optional table can include the following columns: `energy`, `enthalpy`, `entropy`, `free_energy`, `zpe`, and `temperature`. For example:

```
water = Species H2O (
    (energy: -14.22696999) [electron_volt],
    (zpe: 0.558) [eV],
    (entropy: 2.1669e-3) [eV/K],
    (temperature: 298.15) [K]
)
```

The provided input allows to evaluate the free energy at the given temperature. The free energy can be retrieved with `water.free_energy`. It is noted that `water.free_energy` is a series with the length of the series provided in the table in the species definition. To get the first element, `water.free_energy[0]` has to be used. Here an example with two temperatures:


```
water = Species H2O (
    (energy: -14.22696999, -14.22696999) [electron_volt / particle],
    (zpe: 0.558, 0.558) [eV / particle],
    (entropy: 206.5, 213.1) [J/K/mol],
    (temperature: 500., 600.) [K]
)
print(water.free_energy)
print('temperature, free energy:', water.temperature[0], water.free_energy[0])
```

In the latter example, `water.free_energy` includes two elements, for 500 and 600 K, respectively:

```
program output: >>>
(free_energy: -14.739080832009071, -14.994145508249682) [electron_volt / particle]
'temperature, free energy:', 500.0 [kelvin], -14.739080832009071 [electron_volt / particle]
<<<
```

The interpreter evaluates all themochemistry quantities that can be determined by using the available data. For example, if the free energy is given then the energy and the enthalpy will be evaluated:

```
water = Species H2O (
    (free_energy: -14.739080832009071, -14.994145508249682) [electron_volt / particle],
    (zpe: 0.558, 0.558) [eV / particle],
    (entropy: 206.5, 213.1) [J/K/mol],
    (temperature: 500., 600.) [K]
)
print(water.energy, water.enthalpy)
```
Output:
```
program output: >>>
(energy: -14.22696999, -14.22696999) [electron_volt / particle] (enthalpy: -13.66896999, -13.66896999) [electron_volt / particle]
<<<
```


## Chemical reactions

An equation of a chemical reaction is defined with the syntax:

```
Reaction <educts> [=|->] <products> [: table with parameters and properties]
```
The word `Reaction` is a keyword. The individual educts and products must be separated by a `+` sign and must be references to defined species. Every educt and product has a stoichiometric coefficient (a real number). If no coefficient is specified, then `1.0` is assumed. Example:

```
react = Reaction 2 H2 + O2 = 2 H2O: ((free_energy: -4.916) [eV])
```

Thermochemical quantities for a reaction will be evaluated if the necessary data about the species is available. These quantities can be retrieved in the same way as for species. Example:

```
react = Reaction 2 H2 + O2 = 2 H2O

H2O = Species H2O (
    (free_energy: -2.458) [eV],
    (entropy: 2.1669e-3) [eV/K],
    (zpe: 0.558) [eV],
    (temperature: 298.15) [K]
)

H2 = Species H2 (
    (free_energy: 0.0) [eV],
    (zpe: 0.270) [eV],
    (entropy: 1.3613e-3) [eV/K],
    (temperature: 298.15) [K]
)

O2 = Species O2 (
    (free_energy: 0.0) [eV],
    (entropy: 2.1370e-3) [eV/K],
    (zpe: 0.098) [eV],
    (temperature: 298.15) [K]
)

print(react.free_energy)
print(react.enthalpy)
print(react.entropy)
```

```
program output: >>>
(free_energy: -4.916) [electron_volt]
(enthalpy: -5.07276727) [electron_volt]
(entropy: -0.0005258000000000007) [electron_volt / kelvin]
<<<
```