"""main script for model execution in one of three different modes"""
import os
import warnings
from argparse import ArgumentParser
from textx import metamodel_from_file
from textx import get_children_of_type
from fireworks.fw_config import LAUNCHPAD_LOC
from fireworks.scripts.lpad_run import get_lp
from fireworks.user_objects.queue_adapters.common_adapter import CommonAdapter
from virtmat.language.metamodel.properties import add_properties
from virtmat.language.metamodel.processors import add_processors
from virtmat.language.interpreter.workflow_executor import NonCompletedException
from virtmat.language.interpreter.session_manager import Session
from virtmat.language.utilities.errors import error_handler
from virtmat.language.utilities.textx import GRAMMAR_LOC
from virtmat.language.utilities import logging


def parse_clargs():
    """parse the command line arguments for the script"""
    m_description = 'Execute a program in one of three different modes'
    parser = ArgumentParser(description=m_description)
    add_arguments(parser)
    return parser.parse_args()


def add_arguments(parser):
    """add arguments to a parser object"""
    parser.add_argument('-f', '--model_file', required=True,
                        help='path to model file')
    parser.add_argument('-g', '--grammar_path', required=False,
                        help='path to grammar root file', default=GRAMMAR_LOC)
    parser.add_argument('-m', '--mode', required=False, default='instant',
                        help='execution mode', choices=['instant', 'deferred', 'workflow'])
    parser.add_argument('-p', '--print-all-vars', default=False, required=False,
                        action='store_true', help='prints all variables')
    parser.add_argument('-t', '--textx-debug', default=False, required=False,
                        action='store_true', help='enable textx debug mode')
    parser.add_argument('-l', '--launchpad_file', default=LAUNCHPAD_LOC,
                        help='path to launchpad file (workflow mode only)')
    parser.add_argument('-u', '--uuid', default=None, required=False,
                        help='UUID of a model to extend/run (workflow mode only)')
    parser.add_argument('-r', '--autorun', default=False, required=False,
                        action='store_true', help='run the model (workflow mode only)')
    parser.add_argument('-d', '--on-demand', default=False, required=False,
                        action='store_true', help='run on demand (workflow mode only)')
    parser.add_argument('-a', '--async-run', default=False, required=False,
                        action='store_true', help='run in background (workflow mode only)')
    parser.add_argument('-q', '--qadapter_file', default=None, required=False,
                        help='path to default qadapter file (workflow mode only)')
    parser.add_argument('--enable_logging', default=False, required=False,
                        action='store_true', help='enable logging messages')
    parser.add_argument('--logging_level', default='CRITICAL', required=False,
                        choices=logging.LOGGING_LEVELS, help='logging level')


@error_handler
def run_instant_deferred(clargs, deferred=False):
    """run a model in instant or deferred mode"""

    meta = metamodel_from_file(clargs.grammar_path, auto_init_attributes=False,
                               debug=clargs.textx_debug)
    add_properties(meta, deferred_mode=deferred)
    add_processors(meta, constr_processors=True)

    with open(clargs.model_file, 'r', encoding='utf-8') as inp:
        model_str = inp.read()
    return meta.model_from_file(clargs.model_file, deferred_mode=deferred,
                                source_code=model_str, debug=clargs.textx_debug)


@error_handler
def run_workflow(clargs):
    """run a model through a workflow system"""
    config_dir = os.path.dirname(os.path.abspath(clargs.launchpad_file))
    qadapter = CommonAdapter.from_file(clargs.qadapter_file) if clargs.qadapter_file else None
    session = Session(get_lp(clargs), uuid=clargs.uuid, grammar_path=clargs.grammar_path,
                      model_path=clargs.model_file, autorun=clargs.autorun,
                      on_demand=clargs.on_demand, textx_debug=clargs.textx_debug,
                      async_run=clargs.async_run, config_dir=config_dir,
                      qadapter=qadapter)
    prog = session.get_model()
    session.stop_runner()
    if prog:
        print('program UUID:', prog.uuid)
    return prog


@error_handler
def evaluate_prog(clargs, prog):
    """evaluate the model"""
    print(f'program output: >>>\n{prog.value}\n<<<')
    if clargs.print_all_vars:
        for var in get_children_of_type('Variable', prog):
            try:
                print(f'var: {var.name}, type: {var.type_}, value: {var.value}')
            except NonCompletedException:
                print(f'var: {var.name}, type: {var.type_}, value: n.c.')


def main(clargs):
    """main function"""
    logging.LOGGING_LEVEL = logging.get_logging_level(clargs.logging_level)
    if not clargs.enable_logging:
        logging.disable_logging()

    if clargs.mode != 'workflow':

        def check_clargs(attr_, val_, bstr_):
            if ((val_ in (None, False, True) and getattr(clargs, attr_)) or
               getattr(clargs, attr_) != val_):
                warnings.warn(f'argument {attr_} {bstr_}')
        bstr = f'will be ignored in {clargs.mode} mode'
        attrs = {'launchpad_file': LAUNCHPAD_LOC, 'uuid': None, 'on_demand': False,
                 'autorun': False, 'async_run': False, 'qadapter_file': None}
        for attr, val in attrs.items():
            check_clargs(attr, val, bstr)
    if clargs.mode == 'instant':
        prog = run_instant_deferred(clargs, deferred=False)
    elif clargs.mode == 'deferred':
        prog = run_instant_deferred(clargs, deferred=True)
    elif clargs.mode == 'workflow':
        prog = run_workflow(clargs)
    if prog is not None and not isinstance(prog, str):
        evaluate_prog(clargs, prog)


if __name__ == '__main__':
    main(parse_clargs())
