# pylint: disable = protected-access
"""processors to set attributes necessary for the workflow executor"""
import math
from textx import get_children_of_type, get_parent_of_type
from virtmat.language.utilities.serializable import ureg
from virtmat.language.utilities.resources import res_config
from virtmat.language.utilities.dupefinder import DupeFinderFunctionTask
from virtmat.language.constraints.units import check_number_literal_units
from virtmat.language.constraints.values import check_resources_values
from virtmat.language.constraints.values import check_resources_values_limits
from virtmat.language.constraints.values import check_number_of_chunks


def set_category(resources):
    """Set the category to be used for a Firework launch"""
    if resources.walltime is None:
        if resources.ncores is None:
            resources.category = 'interactive'
        else:
            if resources.ncores == 1:
                resources.category = 'interactive'
            else:
                resources.category = 'batch'
    else:
        if resources.walltime.value < ureg.Quantity(1, 'minutes'):
            if resources.ncores is not None and resources.ncores > 1:
                resources.category = 'batch'
            else:
                resources.category = 'interactive'
        else:
            resources.category = 'batch'


def set_queue_adapter(resources):
    """Construct a queue adapter dictionary from a Resources model object. An
    execution model with multiple single-threaded, non-accelerated tasks(ranks)
    with balanced workload over the compute nodes is used"""
    qadapter = {}
    qadapter['queue'] = res_config['queue']
    if resources.walltime is None:
        walltime = res_config['default_walltime']
    else:
        walltime = resources.walltime.value
    qadapter['walltime'] = int(walltime.to('minutes').magnitude)
    if resources.ncores is None:
        ncores = res_config['default_ncores']
    else:
        ncores = resources.ncores
    qadapter['nodes'] = math.ceil(ncores/res_config['max_tasks_per_node'])
    qadapter['ntasks'] = ncores // res_config['default_cpus_per_task']
    qadapter['ntasks_per_node'] = qadapter['ntasks'] // qadapter['nodes']
    if resources.memory is not None:
        qadapter['mem_per_cpu'] = resources.memory.value.to('MB').magnitude
    resources.qadapter = qadapter


def resources_processor(obj):
    """Apply unit constraints to Resources object attributes and then set
    category and qadapter attributes in the Resources object"""
    for attr, unit in (('memory', 'bit'), ('walltime', 'second')):
        number = getattr(obj, attr)
        if number is not None:
            check_number_literal_units(number, attr, unit)
    check_resources_values(obj)
    check_resources_values_limits(obj)
    set_category(obj)
    if obj.category == 'batch':
        set_queue_adapter(obj)
    else:
        obj.qadapter = None


def parallelizable_processor(obj):
    """Processors for parallelizable objects"""
    check_number_of_chunks(obj)


def get_object_str(src, obj):
    """extract the source string of a textx object from the model string"""
    return src[obj._tx_position:obj._tx_position_end].rstrip()


def source_code_statement_processor(model, _):
    """add source code lines pertinent to the variable objects in the model"""
    if not isinstance(model, str):
        p_src = getattr(model, '_tx_model_params').get('source_code')
        for obj in get_children_of_type('Variable', model):
            vartuple_obj = get_parent_of_type('VarTuple', obj)
            var_obj = vartuple_obj if vartuple_obj else obj
            o_src = [get_object_str(p_src, var_obj)] if p_src else []
            setattr(obj, 'source_code', o_src)
            if len(o_src) > 0:
                setattr(obj, 'dupefinder', DupeFinderFunctionTask())
        for statement in ['ObjectImports', 'FunctionDefinition', 'ObjectTo']:
            for obj in get_children_of_type(statement, model):
                o_src = [get_object_str(p_src, obj)] if p_src else []
                setattr(obj, 'source_code', o_src)
