"""Custom classes for the AMML objects"""
import types
import math
import numpy
import pandas
import pint
import pint_pandas
import ase
from pint_pandas import PintType
from ase import constraints
from ase.io import write, jsonio
from ase.geometry import wrap_positions
from ase.calculators.calculator import get_calculator_class
from virtmat.language.utilities.errors import RuntimeValueError
from virtmat.language.utilities.errors import ConvergenceError
from virtmat.language.utilities.errors import InvalidUnitError
from virtmat.language.utilities.errors import StructureInputError
from virtmat.language.utilities.units import ureg, calc_pars
from virtmat.language.utilities.lists import list_apply, list_flatten


def get_par_units(name, par, val):
    """auxiliary function to get the units of a calculator parameter"""
    params = calc_pars[name]
    try:
        par_def = params[par]
    except KeyError as err:
        msg = f'parameter {par} has unknown units'
        raise InvalidUnitError(msg) from err
    if isinstance(par_def, types.FunctionType):
        return par_def(val)
    return par_def


def get_params_units(calc_name, params):
    """auxiliary function to get the units for a parameters dictionary"""
    units = {}
    for param, value in params.items():
        r_units = get_par_units(calc_name, param, value)
        if isinstance(r_units, list):
            r_units = [r(params) for r in r_units]
            r_units = next(r for r in r_units if r is not None)
        units[param] = r_units
    return units


def get_params_magnitudes(calc_params, units):
    """converts parameters to canonical units and return their magnitudes"""

    def magnitudes_from_list(unit, value):
        return list_apply(lambda x: x.to(unit).magnitude, value)

    magnitudes = {}
    for par, val in calc_params.items():
        try:
            if isinstance(val, ureg.Quantity):
                magnitudes[par] = val.to(units[par]).magnitude
            elif isinstance(val, numpy.ndarray):
                if issubclass(val.dtype.type, (numpy.bool_, numpy.str_)):
                    magnitudes[par] = val
                else:
                    assert issubclass(val.dtype.type, numpy.object_)
                    magnitudes[par] = magnitudes_from_list(units[par], val.tolist())
            else:
                assert isinstance(val, (bool, str)) or val is None
                magnitudes[par] = val
        except pint.DimensionalityError as err:
            msg = (f'error with units of parameter \"{par}\": '
                   f'must be [{units[par]}] instead of [{val.units}]')
            raise InvalidUnitError(msg) from err
    return magnitudes


def get_structure_dataframe(atoms):
    """create a structure dataframe from an ASE Atoms object"""
    lunit = pint_pandas.PintType('angstrom')
    munit = pint_pandas.PintType('amu')
    punit = pint_pandas.PintType('amu * angstrom / fs')

    symb = pandas.Series(atoms.get_chemical_symbols(), name='symbols')
    post = atoms.get_positions().transpose()
    posx = pandas.Series(pint_pandas.PintArray(post[0], dtype=lunit), name='x')
    posy = pandas.Series(pint_pandas.PintArray(post[1], dtype=lunit), name='y')
    posz = pandas.Series(pint_pandas.PintArray(post[2], dtype=lunit), name='z')
    momt = atoms.get_momenta().transpose()
    momx = pandas.Series(pint_pandas.PintArray(momt[0], dtype=punit), name='px')
    momy = pandas.Series(pint_pandas.PintArray(momt[1], dtype=punit), name='py')
    momz = pandas.Series(pint_pandas.PintArray(momt[2], dtype=punit), name='pz')
    tags = pandas.Series(atoms.get_tags(), name='tags')
    mass = pandas.Series(pint_pandas.PintArray(atoms.get_masses(), dtype=munit), name='masses')

    atoms_cols = [symb, posx, posy, posz, momx, momy, momz, mass, tags]
    atoms_df = pandas.concat(atoms_cols, axis=1)
    cell = ureg.Quantity(atoms.get_cell().array, 'angstrom')
    struct_cols = [{'atoms': atoms_df, 'cell': cell, 'pbc': atoms.get_pbc()}]
    return pandas.DataFrame(struct_cols)


def get_calculator_dataframe(ase_calc):
    """create a calculator dataframe from an ASE Calculator object"""
    parameters = jsonio.decode(jsonio.encode(ase_calc.parameters))
    units = get_params_units(ase_calc.name.lower(), parameters)
    calc_params = {}
    for par, val in parameters.items():
        if isinstance(val, (bool, str)) or val is None:
            calc_params[par] = val
        elif isinstance(val, (int, float)):
            calc_params[par] = ureg.Quantity(val, units[par])
        elif isinstance(val, (tuple, list)):
            if all(isinstance(v, (bool, str)) for v in list_flatten(val)):
                calc_params[par] = numpy.array(val)
            else:
                assert all(isinstance(v, (int, float)) for v in list_flatten(val))
                assert units[par] is not None
                calc_params[par] = ureg.Quantity(numpy.array(val), units[par])
        elif isinstance(val, numpy.ndarray):
            if numpy.issubdtype(val.dtype, numpy.number):
                assert units[par] is not None
                calc_params[par] = ureg.Quantity(val, units[par])
            else:
                npdtypes = (numpy.bool_, numpy.str_)
                assert any(numpy.issubdtype(val.dtype, t) for t in npdtypes)
                assert units[par] is None
                calc_params[par] = val
        else:
            raise TypeError('unknown type in calculator parameters')
    return pandas.DataFrame([calc_params])


def get_vibrational_energies(calc):
    """an interface to return vibrational energies"""
    if calc.name.lower() == 'vasp':
        (vib_ene_real, vib_ene_imag) = calc.read_vib_freq()  # returns in meV
        vib_ene_real = numpy.array(vib_ene_real)*1.0e-3
        vib_ene_imag = numpy.array(vib_ene_imag)*1.0e-3
    elif calc.name.lower() == 'turbomole':
        vib_ene_real = []
        vib_ene_imag = []
        for eigenvalue in calc.get_results()['vibrational spectrum']:
            if eigenvalue['irreducible representation']:
                assert eigenvalue['frequency']['units'] == 'cm^-1'
                fact = ase.units.invcm
                if eigenvalue['frequency']['value'] > 0.0:
                    vib_ene_real.append(eigenvalue['frequency']['value']*fact)
                else:
                    vib_ene_imag.append(eigenvalue['frequency']['value']*fact)
    else:
        msg = f'vibrational energies not available for calculator {calc.name}'
        raise NotImplementedError(msg)
    return vib_ene_real, vib_ene_imag


def tm_dipole_handler(lst):
    """handler function for the electric dipole moment"""
    res = []
    for dct in lst:
        magnitude = dct['absolute value']['value']
        units = dct['absolute value']['units'].lower()
        assert units == 'debye'
        ref = ureg.Quantity(magnitude, units)
        assert dct['vector']['units'] == 'a.u.'
        vec = ureg.Quantity(dct['vector']['array'], EC_BOHR)
        test = pint.Quantity(numpy.linalg.norm(vec.to('D').magnitude), vec.to('D').units)
        assert math.isclose(test.magnitude, ref.magnitude, rel_tol=0.001)
        res.append(vec)
    return pandas.Series(res)


def ase_vibrational_energies_handler(vib):
    """handler function for the vibrational energies in ASE units [eV]"""
    dtype = pint_pandas.PintType('eV')
    data = (pandas.Series(re, name='vibrational_energies', dtype=dtype) for re, im in vib)
    return pandas.Series(data)


EC_ANG = 'elementary_charge * angstrom'
EC_BOHR = 'elementary_charge * bohr'
EV_PER_ANG = 'eV / angstrom'

ase_p_df = pandas.DataFrame(
    [
        {
         'calc_name': 'lj',
         'prop_name': 'energy',
         'internal_name': 'energy',
         'handler': lambda x: pandas.Series(x, dtype=PintType('eV'))
        },
        {
         'calc_name': 'lj',
         'prop_name': 'forces',
         'internal_name': 'forces',
         'handler': lambda x: pandas.Series(ureg.Quantity(i, EV_PER_ANG) for i in x)
        },
        {
         'calc_name': 'emt',
         'prop_name': 'energy',
         'internal_name': 'energy',
         'handler': lambda x: pandas.Series(x, dtype=PintType('eV'))
        },
        {
         'calc_name': 'emt',
         'prop_name': 'forces',
         'internal_name': 'forces',
         'handler': lambda x: pandas.Series(ureg.Quantity(i, EV_PER_ANG) for i in x)
        },
        {
         'calc_name': 'vasp',
         'prop_name': 'energy',
         'internal_name': 'energy',
         'handler': lambda x: pandas.Series(x, dtype=PintType('eV'))
        },
        {
         'calc_name': 'vasp',
         'prop_name': 'dipole',
         'internal_name': 'dipole',
         'handler': lambda x: pandas.Series(ureg.Quantity(i, EC_ANG) for i in x)
        },
        {
         'calc_name': 'vasp',
         'prop_name': 'forces',
         'internal_name': 'forces',
         'handler': lambda x: pandas.Series(ureg.Quantity(i, EV_PER_ANG) for i in x)
        },
        {
         'calc_name': 'vasp',
         'prop_name': 'vibrational_energies',
         'internal_name': 'vibrational_energies',
         'handler': ase_vibrational_energies_handler
        },
        {
         'calc_name': 'vasp',
         'prop_name': 'energy_minimum',
         'internal_name': 'vibrational_energies',
         'handler': lambda x: pandas.Series(len(im) == 0 for re, im in x)
        },
        {
         'calc_name': 'vasp',
         'prop_name': 'transition state',
         'internal_name': 'vibrational_energies',
         'handler': lambda x: pandas.Series(len(im) == 1 for re, im in x)
        },
        {
         'calc_name': 'turbomole',
         'prop_name': 'energy',
         'internal_name': 'total energy',
         'handler': lambda x: pandas.Series(x, dtype=PintType('eV'))
        },
        {
         'calc_name': 'turbomole',
         'prop_name': 'dipole',
         'internal_name': 'electric dipole moment',
         'handler': tm_dipole_handler,
        },
        {
         'calc_name': 'turbomole',
         'prop_name': 'forces',
         'internal_name': 'energy gradient',
         'handler': lambda x: pandas.Series(-ureg.Quantity(i, EV_PER_ANG) for i in x)
        },
        {
         'calc_name': 'turbomole',
         'prop_name': 'vibrational_energies',
         'internal_name': 'vibrational_energies',
         'handler': ase_vibrational_energies_handler
        },
        {
         'calc_name': 'turbomole',
         'prop_name': 'energy_minimum',
         'internal_name': 'vibrational_energies',
         'handler': lambda x: pandas.Series(len(im) == 0 for re, im in x)
        },
        {
         'calc_name': 'turbomole',
         'prop_name': 'transition_state',
         'internal_name': 'vibrational_energies',
         'handler': lambda x: pandas.Series(len(im) == 1 for re, im in x)
        },
     ]
)


def get_ase_property(calc_name, prop_name, results):
    """convert a list of dictionaries into series of a specified property"""
    dfr = ase_p_df[(ase_p_df['calc_name'] == calc_name) & (ase_p_df['prop_name'] == prop_name)]
    if len(dfr['internal_name']) == 0:
        msg = f'no property {prop_name} found for calculator {calc_name}'
        raise RuntimeError(msg)
    assert len(dfr['internal_name']) == 1
    internal_name = next(iter(dfr['internal_name']))
    handler_func = next(iter(dfr['handler']))
    return handler_func([r[internal_name] for r in results])


class AMMLObject:
    """base class for all AMML objects"""


class AMMLStructure(AMMLObject):
    """"custom AMML Structure class"""
    def __init__(self, tab, name=None):
        if 'pbc' in tab and 'cell' not in tab:
            for pbc in tab['pbc'].values:
                if any(pbc):
                    raise RuntimeValueError('cell must be specified with pbc')
        self.name = name
        self.tab = tab

    def __getitem__(self, key):
        if isinstance(key, int):
            dfr = self.tab.iloc[[key]]
            return tuple(next(dfr.itertuples(index=False, name=None)))
        if isinstance(key, str):
            if key == 'name':
                return self.name
            return self.tab[key]
        if isinstance(key, slice):
            return self.__class__(self.tab[key], self.name)
        raise TypeError('unknown key type')

    def to_ase_file(self, filename):
        """store series of ase.Atoms objects to a file in supported format"""
        atoms_objs = self.to_ase()
        write(images=atoms_objs, filename=filename)

    def to_ase(self):
        """create a series of ase.Atoms objects from a structure dataframe"""
        dfr_ase = self.get_ase_atoms_dataframe()
        series = dfr_ase.apply(lambda row: ase.Atoms(**row.to_dict()), axis=1)
        return series.rename(self.name, inplace=True)

    def get_ase_atoms_dataframe(self):
        """return a unitless atoms dataframe from a structure dataframe with units"""
        dfr = self.tab
        dfr_ase = pandas.DataFrame()
        dfr_ase['symbols'] = dfr['atoms'].apply(lambda x: x.symbols.to_numpy())
        if 'cell' in dfr:
            dfr_ase['cell'] = dfr['cell'].apply(lambda x: x.to('angstrom').magnitude)
        else:
            dfr_ase[['cell']] = None

        dfr_ase['pbc'] = dfr['pbc'].to_numpy() if 'pbc' in dfr else None

        def xyz2pos(atoms):
            xcoord = atoms['x'].pint.to('angstrom').values.data
            ycoord = atoms['y'].pint.to('angstrom').values.data
            zcoord = atoms['z'].pint.to('angstrom').values.data
            return numpy.array([xcoord, ycoord, zcoord]).transpose()

        def pxpypz2momenta(atoms):
            if 'px' not in atoms:
                return None
            units = 'amu * angstrom / fs'
            xcoord = atoms['px'].pint.to(units).values.data / ase.units.fs
            ycoord = atoms['py'].pint.to(units).values.data / ase.units.fs
            zcoord = atoms['pz'].pint.to(units).values.data / ase.units.fs
            return numpy.array([xcoord, ycoord, zcoord]).transpose()

        dfr_ase['positions'] = dfr['atoms'].apply(xyz2pos)
        dfr_ase['momenta'] = dfr['atoms'].apply(pxpypz2momenta)
        dfr_ase['tags'] = dfr['atoms'].apply(lambda x: x.tags.to_numpy() if 'tags' in x else None)
        return dfr_ase

    @classmethod
    def from_ase(cls, atoms, name=None):
        """create an AMML object from an ASE Atoms object"""
        if isinstance(atoms, ase.Atoms):
            return cls(get_structure_dataframe(atoms), name)
        assert isinstance(atoms, (list, tuple, pandas.Series))
        atoms_dfs = [get_structure_dataframe(a) for a in atoms]
        return cls(pandas.concat(atoms_dfs, ignore_index=True), name)

    @classmethod
    def from_ase_file(cls, filename):
        """create an AMML Structure object from a file in an ASE-supported format"""
        try:
            structs = ase.io.read(filename, index=':')
        except Exception as err:  # broad exception due to ase.io.read
            msg = f'{err.__class__.__name__}: {str(err)}'
            raise StructureInputError(msg) from err
        struct_dfs = [get_structure_dataframe(s) for s in structs]
        return cls(pandas.concat(struct_dfs, ignore_index=True))


class Calculator(AMMLObject):
    """custom AMML Calculator class"""
    def __init__(self, name, parameters, pinning=None, version=None):
        self.name = name
        self.parameters = parameters
        self.pinning = pinning
        self.version = version

    def __getitem__(self, key):
        if isinstance(key, int):
            dfr = self.parameters.iloc[[key]]
            return tuple(next(dfr.itertuples(index=False, name=None)))
        if isinstance(key, str):
            return getattr(self, key)
        if isinstance(key, slice):
            return self.__class__(self.name, self.parameters[key], self.pinning,
                                  self.version)
        raise TypeError('unknown key type')

    def to_ase(self):
        """create a series of ASE Calculator objects"""
        dfr = pandas.DataFrame() if self.parameters is None else self.parameters
        calc_list = []
        for _, row in dfr.iterrows():
            calc_params = dict(row)
            units = get_params_units(self.name, calc_params)
            calc_class = get_calculator_class(self.name)
            calc_list.append(calc_class(**get_params_magnitudes(calc_params, units)))
        return pandas.Series(calc_list, name='calc')

    @classmethod
    def from_ase(cls, ase_calc):
        """create an AMML Calculator object from an ASE Calculator object"""
        return cls(ase_calc.name.lower(), get_calculator_dataframe(ase_calc))


class Property(AMMLObject):
    """custom AMML Property class"""
    def __init__(self, names, calc, struct, constrs=None, task=None, results=None):
        self.names = names
        self.calc = calc
        self.struct = struct
        if constrs:
            if not all(len(c.fixed) == len(a) for c in constrs for a in struct.tab.atoms):
                msg = ('The list of fixed/non-fixed atoms in constraints and '
                       'atoms in structure have different lengths')
                raise RuntimeValueError(msg)
            self.constrs = constrs
        else:
            self.constrs = []
        self.task = task
        if self.task is not None:
            self.requires_dof = not self.task == 'single point'
        else:
            self.requires_dof = None  # fixme: check the calc params and algorithm
        if results is None:
            self.dof_vector, self.dof_number = self.get_dof()
            if self.requires_dof and self.dof_number == 0:
                msg = 'All degrees of freedom frozen. Hint: check task and constraints.'
                raise RuntimeValueError(msg)
            self.results = self.get_results()
        else:
            self.results = results

    def __getitem__(self, key):
        if isinstance(key, int):
            dfr = self.results.iloc[[key]]
            return tuple(next(dfr.itertuples(index=False, name=None)))
        if isinstance(key, str):
            if key in ('names', 'calc', 'struct', 'constrs', 'task', 'results'):
                return getattr(self, key)
            return getattr(self.results, key)
        if isinstance(key, slice):
            struct_dfs = [s.tab for s in self.results.struct[key]]
            struct_df = pandas.concat(struct_dfs, ignore_index=True)
            struct_name = self.results.struct[key][0].name
            struct = AMMLStructure(struct_df, struct_name)
            calc_dfs = [c.parameters for c in self.results.calc[key]]
            calc_df = pandas.concat(calc_dfs, ignore_index=True)
            calc_name = self.results.calc[key][0].name
            calc_pinning = self.results.calc[key][0].pinning
            calc_version = self.results.calc[key][0].version
            calc = Calculator(calc_name, calc_df, calc_pinning, calc_version)
            return self.__class__(self.names, calc, struct, constrs=self.constrs,
                                  task=self.task, results=self.results[key])
        raise TypeError('unknown key type')

    def get_dof(self):
        """return the non-frozen nuclear degrees of freedom"""
        atoms = self.struct[0:1].to_ase()[0]
        if self.constrs:
            new_pos = numpy.array(atoms.positions)
            for ind, (pbc, vec) in enumerate(zip(atoms.pbc, atoms.cell)):
                if pbc:
                    new_pos += 0.9*vec
                else:
                    new_pos[:, ind] += 0.9
            wrap_positions(new_pos, atoms.cell, atoms.pbc)
            constrs = []
            for constr in self.constrs:
                constrs.extend(constr.to_ase())
            for constr in constrs:
                constr.adjust_positions(atoms, new_pos)
            epsilon = numpy.finfo(numpy.float64).eps
            dof = numpy.abs(atoms.positions-new_pos) > epsilon
        else:
            dof = numpy.full((len(atoms), 3), True)
        return dof, numpy.sum(dof)

    def get_results(self):
        """create a dataframe with properties calculated using ASE"""
        props = self.names
        calcs = self.calc.to_ase()
        structs = self.struct.to_ase()
        props_dict = {}
        for name in props + ['calc', 'struct', 'output_structure', 'results']:
            props_dict[name] = []
        for calc in calcs:
            for struct in structs:
                props_dict['calc'].append(Calculator.from_ase(calc))
                props_dict['struct'].append(AMMLStructure.from_ase(struct, self.struct.name))
                n_struct = ase.Atoms(struct)
                constrs = []
                for constr in self.constrs:
                    constrs.extend(constr.to_ase())
                n_struct.constraints = constrs
                calc.calculate(n_struct)
                if hasattr(calc, 'converged'):
                    if not calc.converged:
                        msg = f'calculation of {", ".join(props)} not converged'
                        raise ConvergenceError(msg)
                results = calc.results
                if self.task == 'normal modes':
                    results['vibrational_energies'] = get_vibrational_energies(calc)
                props_dict['results'].append(results)
                n_struct.calc = None
                struct_update = AMMLStructure.from_ase(n_struct, self.struct.name)
                props_dict['output_structure'].append(struct_update)
        for prop in props:
            props_dict[prop] = get_ase_property(self.calc.name, prop, props_dict['results'])
        del props_dict['results']
        props_dict['calc'] = pandas.Series(props_dict['calc'])
        props_dict['struct'] = pandas.Series(props_dict['struct'])
        props_dict['output_structure'] = pandas.Series(props_dict['output_structure'])
        return pandas.DataFrame(props_dict)


class Constraint(AMMLObject):
    """custom AMML Constraint class"""
    def __init__(self, name, **kwargs):
        assert name in ['FixedAtoms', 'FixedLine', 'FixedPlane']
        self.name = name
        self.kwargs = kwargs
        self.fixed = kwargs['fixed']
        self.indices = self.fixed[self.fixed].index.values
        if name in ['FixedLine', 'FixedPlane']:
            self.direction = kwargs['direction']

    def to_ase(self):
        """return ASE constraint objects"""
        if self.name == 'FixedAtoms':
            return [constraints.FixAtoms(indices=self.indices)]
        if self.name == 'FixedLine':
            cls = constraints.FixedLine
        elif self.name == 'FixedPlane':
            cls = constraints.FixedPlane
        direction = [c.magnitude for c in self.direction.tolist()]
        return [cls(i, direction=direction) for i in self.indices]


class NeighborList(AMMLObject):
    """custom AMML NeighborList class"""
