"""helper functions to interface fireworks"""
from fireworks import FWorker
from fireworks.core.rocket_launcher import rapidfire, launch_rocket


def run_fireworks(lpad, fw_ids, create_subdirs=False):
    """launch fireworks with the provided fw_ids"""
    worker = FWorker(name='local', query={'fw_id': {'$in': fw_ids}})
    if create_subdirs:
        rapidfire(lpad, fworker=worker, nlaunches=len(fw_ids), strm_lvl=lpad.strm_lvl)
    for _ in fw_ids:
        launch_rocket(lpad, fworker=worker, strm_lvl=lpad.strm_lvl)


def get_nodes_providing(lpad, fw_ids, name):
    """find all nodes providing a specified output in the list of fw_ids"""
    fw_query = {'fw_id': {'$in': fw_ids}, 'spec._tasks.outputs': name}
    return lpad.get_fw_ids(fw_query)


def get_root_node_ids(lpad, fw_ids):
    """find all root nodes in the list of fw_ids"""
    return lpad.get_fw_ids({'fw_id': {'$in': fw_ids}, 'name': '_fw_root_node'})


def get_parent_nodes(lpad, fw_ids, firework):
    """find parent nodes for a new firework in the list of fw_ids
       The fw_ids must belong to one workflow that must have one root node."""
    if all('inputs' in t and len(t['inputs']) == 0 for t in firework.tasks):
        parents = get_root_node_ids(lpad, fw_ids)
        assert len(parents) == 1
    else:
        inputs = [i for t in firework.tasks for i in t['inputs']]
        parents = set()
        for inp in inputs:
            nodes = get_nodes_providing(lpad, fw_ids, inp)
            if len(nodes) != 0:
                assert len(nodes) == 1
                parents.update(nodes)
            elif inp in firework.spec:
                parents.add('root_node')
            else:
                parents.add(None)
        assert len(parents) > 0
        if 'root_node' in parents:
            parents.remove('root_node')
            if len(parents) == 0:
                parents = get_root_node_ids(lpad, fw_ids)
                assert len(parents) == 1
    return list(parents)


def get_ancestors(lpad, fw_id):
    """return a list of all ancestors' fw_ids of a node with fw_id"""
    # better would be a direct pymongo query with a projection on links
    wfl_pl = lpad.get_wf_by_fw_id(fw_id).links.parent_links
    parents = wfl_pl.get(fw_id, [])
    ancestors = set()
    while parents:
        ancestors.update(parents)
        new_parents = set()
        for par in iter(parents):
            new_parents.update(wfl_pl.get(par, []))
        parents = new_parents
    return list(ancestors)


def safe_update(lpad, fw_id, update_dict):
    """update the node spec in a safe way"""
    state = lpad.get_fw_by_id(fw_id).state
    if state == 'COMPLETED':
        lpad.defuse_fw(fw_id)
        lpad.update_spec([fw_id], update_dict)
        lpad.reignite_fw(fw_id)
    elif state in ['WAITING', 'READY']:
        lpad.update_spec([fw_id], update_dict)
    elif state == 'FIZZLED':
        lpad.update_spec([fw_id], update_dict)
        lpad.rerun_fw(fw_id)


def get_one_node_info(lpad, wf_query, fw_query, fw_proj, wf_proj=None):
    """find one node for each matching workflow and returns selected info"""
    wf_proj = {'metadata.uuid': True} if wf_proj is None else wf_proj
    fw_pipeline = [{'$match': {**fw_query, '$expr': {'$in': ['$fw_id', '$$mynodes']}}},
                   {'$project': {**fw_proj, '_id': False}}]
    wf_lookup = {'from': 'fireworks', 'let': {'mynodes': '$nodes'},
                 'pipeline': fw_pipeline, 'as': 'nodes'}
    wf_pipeline = [{'$match': wf_query}, {'$lookup': wf_lookup},
                   {'$addFields': {'node': {'$arrayElemAt': ['$nodes', 0]}}},
                   {'$project': {'_id': False, 'node': True, **wf_proj}}]
    cursor = lpad.workflows.aggregate(wf_pipeline)
    return list(cursor)


def get_all_nodes_info(lpad, wf_query, fw_query, fw_proj, wf_proj=None):
    """find all nodes for each matching workflow and returns selected info"""
    wf_proj = {'metadata.uuid': True} if wf_proj is None else wf_proj
    fw_pipeline = [{'$match': {**fw_query, '$expr': {'$in': ['$fw_id', '$$mynodes']}}},
                   {'$project': {**fw_proj, '_id': False}}]
    wf_lookup = {'from': 'fireworks', 'let': {'mynodes': '$nodes'},
                 'pipeline': fw_pipeline, 'as': 'nodes'}
    wf_pipeline = [{'$match': wf_query}, {'$lookup': wf_lookup},
                   {'$project': {'_id': False, 'nodes': True, **wf_proj}}]
    cursor = lpad.workflows.aggregate(wf_pipeline)
    return list(cursor)
