"""
tests for amml data structures and operations
"""
import os
import pytest
from textx import get_children_of_type
from textx.exceptions import TextXError
from virtmat.language.utilities.types import typemap


@pytest.fixture(name='water_yaml')
def water_yaml_fixture(tmp_path):
    """path to water.yaml file"""
    return os.path.join(tmp_path, 'water.yaml')


@pytest.fixture(name='water_cif')
def water_cif_fixture(tmp_path):
    """path to water.cif file"""
    return os.path.join(tmp_path, 'water.cif')


@pytest.fixture(name='calc_yaml')
def water_calc_fixture(tmp_path):
    """path to calc.yaml file"""
    return os.path.join(tmp_path, 'calc.yaml')


def test_amml_structure_literal(meta_model, model_kwargs, water_yaml, water_cif):
    """test AMML structure literal and I/O"""
    if 'model_instance' in model_kwargs:
        pytest.skip('test does not work in workflow mode (I/O race condition)')

    inp = ("water = Structure from ("
           "          (atoms: ((symbols: 'O', 'H', 'H'),"
           "                   (x: 0., 0., 0.) [nm],"
           "                   (y: 0., 0.763239, -0.763239) [angstrom],"
           "                   (z: 0.119262, -0.477047, -0.477047) [angstrom],"
           "                   (tags: 1, 0, 0),"
           "                   (masses: 16., 1., 1.) [amu]"
           "                  )"
           "          ),"
           "          (cell: [[2., 0., 0.], [0., 2., 0.], [0., 0., 2.]] [bohr]),"
           "          (pbc: [false, false, true])"
           "        )\n"
           "water to file \'" + water_yaml + "\'\n"
           "water to file \'" + water_cif + "\'\n"
           "water_1 = Structure from file \'" + water_yaml + "\'\n"
           "water_2 = Structure from file \'" + water_cif + "\'")
    prog = meta_model.model_from_str(inp, **model_kwargs)
    var_list = get_children_of_type('Variable', prog)
    water_var = next(v for v in var_list if v.name == 'water')
    assert issubclass(water_var.type_, typemap['AMMLStructure'])
    assert isinstance(water_var.value, typemap['AMMLStructure'])
    water1_var = next(v for v in var_list if v.name == 'water_1')
    water2_var = next(v for v in var_list if v.name == 'water_2')
    assert water2_var.value['atoms'][0].symbols.tolist() == ['O', 'H', 'H']
    assert water1_var.value.name == water_var.value.name
    tab = water1_var.value.tab
    tabref = water_var.value.tab
    assert tab.atoms[0].symbols.tolist() == tabref.atoms[0].symbols.tolist()
    assert tab.atoms[0].x.tolist() == tabref.atoms[0].x.tolist()
    assert tab.atoms[0].y.tolist() == tabref.atoms[0].y.tolist()
    assert tab.atoms[0].z.tolist() == tabref.atoms[0].z.tolist()
    assert tab.atoms[0].tags.tolist() == tabref.atoms[0].tags.tolist()
    assert tab.atoms[0].masses.tolist() == tabref.atoms[0].masses.tolist()
    assert tab.cell[0].data.tolist() == tabref.cell[0].data.tolist()
    assert tab.pbc[0].data.tolist() == tabref.pbc[0].data.tolist()


def test_amml_calculator_literal(meta_model, model_kwargs, calc_yaml):
    """test AMML calculator literal and I/O"""
    inp = ("calc = Calculator vasp ("
           "     (algo: 'Fast'),"
           "     (ediff: 1e-06) [eV],"
           "     (ediffg: -0.01) [eV/angstrom],"
           "     (encut: 400.0) [eV],"
           "     (ibrion: 2),"
           "     (icharg: 2),"
           "     (isif: 2),"
           "     (ismear: 0),"
           "     (ispin: 2),"
           "     (istart: 0),"
           "     (kpts: [5, 5, 1]),"
           "     (lcharg: false),"
           "     (lreal: 'Auto'),"
           "     (lwave: false),"
           "     (nelm: 250),"
           "     (nsw: 1500),"
           "     (potim: 0.1),"
           "     (prec: 'Normal'),"
           "     (sigma: 0.1) [eV],"
           "     (xc: 'PBE')"
           ");"
           "calc to file \'" + calc_yaml + "\';"
           "new_calc = Calculator from file \'" + calc_yaml + "\'")
    prog = meta_model.model_from_str(inp, **model_kwargs)
    var_list = get_children_of_type('Variable', prog)
    calc_var = next(v for v in var_list if v.name == 'calc')
    assert issubclass(calc_var.type_, typemap['AMMLCalculator'])
    assert isinstance(calc_var.value, typemap['AMMLCalculator'])


def test_amml_property_literal(meta_model, model_kwargs):
    """test AMML property literal and I/O"""
    inp = ("h2o = Structure water from ("
           "        (atoms: ((symbols: 'O', 'H', 'H'),"
           "                 (x: 0., 0., 0.) [nm],"
           "                 (y: 0., 0.763239, -0.763239) [angstrom],"
           "                 (z: 0.119262, -0.477047, -0.477047) [angstrom]"
           "                )"
           "        ),"
           "        (cell: [[10., 0., 0.], [0., 10., 0.], [0., 0., 10.]] [angstrom]),"
           "        (pbc: [true, true, true])"
           "      );"
           "calc = Calculator vasp >= 5.4.4 ("
           "          (algo: 'Fast'),"
           "          (ediff: 1e-06) [eV],"
           "          (ediffg: -0.005) [eV/angstrom],"
           "          (encut: 400.0) [eV],"
           "          (ibrion: 2),"
           "          (icharg: 2),"
           "          (isif: 2),"
           "          (ismear: 0),"
           "          (ispin: 2),"
           "          (istart: 0),"
           "          (kpts: [5, 5, 1]),"
           "          (lcharg: false),"
           "          (lreal: 'Auto'),"
           "          (lwave: false),"
           "          (nelm: 250),"
           "          (nsw: 1500),"
           "          (potim: 0.1),"
           "          (prec: 'Normal'),"
           "          (sigma: 0.1) [eV],"
           "          (xc: 'PBE')"
           "       );"
           "props = Property energy, forces for structure h2o with calculator calc")
    meta_model.model_from_str(inp, **model_kwargs)


def test_amml_property_access_properties(meta_model, model_kwargs):
    """test access to properties in an evaluated AMML property literal"""
    inp = ("epsilon_K = 119.8 [K];"
           "kB = 1. [boltzmann_constant];"
           "calc = Calculator lj ((sigma: 3.405) [angstrom], (epsilon: epsilon_K*kB));"
           "struct = Structure fcc_Ar4 from ("
           "     (atoms: ("
           "       (symbols: 'Ar', 'Ar', 'Ar', 'Ar'),"
           "       (x: 0., 2.41, 2.41, 0.) [angstrom],"
           "       (y: 0., 2.41, 0., 2.41) [angstrom],"
           "       (z: 0., 0., 2.41, 2.41) [angstrom]"
           "      )"
           "     ),"
           "     (pbc: [true, true, true]),"
           "     (cell: [[4.82, 0., 0.], [0., 4.82, 0.], [0., 0., 4.82]] [angstrom])"
           ");"
           "prop = Property energy, forces with calculator calc for structure struct;"
           "print(calc, struct);"
           "print(prop);"
           "print(prop.energy);"
           "print(prop.forces); energy = prop.energy[0]")
    prog = meta_model.model_from_str(inp, **model_kwargs)
    print(prog.value)
    var_list = get_children_of_type('Variable', prog)
    energy = next(v for v in var_list if v.name == 'energy')
    assert issubclass(energy.type_, typemap['Quantity'])
    assert isinstance(energy.value, typemap['Quantity'])
    assert energy.value.units == 'electron_volt'
    assert energy.value.magnitude == pytest.approx(-0.16027557460638842)


def test_amml_property_slicing(meta_model, model_kwargs):
    """test amml property slicing"""
    inp = ("h2o = Structure water from ("
           "     (atoms: ((symbols: 'O', 'H', 'H'),"
           "       (x: 0., 0., 0.) [nm],"
           "       (y: 0., 0.763239, -0.763239) [angstrom],"
           "       (z: 0.119262, -0.477047, -0.477047) [angstrom]"
           "      )"
           "     )"
           ");"
           "calc = Calculator emt ((restart: default), (fixed_cutoff: default));"
           "props = Property energy, forces for structure h2o with calculator calc,"
           "                 task: single point;"
           "print(props[0]);"
           "print(props.calc[0:1]);"
           "props_01 = props[0:1];"
           "print(props_01);"
           "print(props_01.struct);"
           "print(props_01.calc)\n")
    prog = meta_model.model_from_str(inp, **model_kwargs)
    print(prog.value)


def test_amml_constraints_properties(meta_model, model_kwargs):
    """test AMML property with AMML calculator and AMML constraints"""
    inp = ("h2o = Structure H2O from ("
           "          (atoms: ((symbols: 'O', 'H', 'H'),"
           "                   (x: 6., 6.96504783, 5.87761163) [angstrom],"
           "                   (y: 6., 5.87761163, 6.96504783) [angstrom],"
           "                   (z: 6., 6.00000000, 6.00000000) [angstrom]"
           "                  )"
           "          ),"
           "          (cell: [[12., 0., 0.], [0., 12., 0.], [0., 0., 12.]] [angstrom]),"
           "          (pbc: [true, true, true])"
           ");"
           "h2 = Structure H2 from ("
           "          (atoms: ((symbols: 'H', 'H'),"
           "               (x: 0., 0.) [angstrom],"
           "               (y: 0., 0.) [angstrom],"
           "               (z: 0., 1.) [angstrom]"
           "            )"
           "          ),"
           "          (cell: [[12., 0., 0.], [0., 12., 0.], [0., 0., 12.]] [angstrom]),"
           "          (pbc: [true, true, true])"
           ");"
           "calc = Calculator emt(default);"
           "calc2 = Calculator emt((fixed_cutoff: default, false));"
           "plane = FixedPlane normal to [0, 0, 1] where (fix: true, true, true);"
           "props = Property energy, forces for structure h2o with calculator calc"
           "                 with constraints (plane), task: single point;"
           "props2 = Property energy, forces for structure h2o"
           "                  with calculator calc2 with constraints (plane);"
           "print(plane);"
           "print(props.struct.name, props.energy);"
           "print(props2.struct.name, props2.energy)")
    prog = meta_model.model_from_str(inp, **model_kwargs)
    print(prog.value)


def test_amml_property_inconsistent_constraints(meta_model, model_kwargs):
    """test AMML property with AMML structure and inconsistent AMML constraints"""
    inp = ("h2 = Structure H2 from ("
           "          (atoms: ((symbols: 'H', 'H'),"
           "               (x: 0., 0.) [angstrom],"
           "               (y: 0., 0.) [angstrom],"
           "               (z: 0., 1.) [angstrom]"
           "            )"
           "          )"
           ");"
           "calc = Calculator emt(default);"
           "constr = FixedAtoms where (fix: true, false, false);"
           "props = Property energy for structure h2 with calculator calc"
           "                 with constraints (constr)\n")
    msg = ('The list of fixed/non-fixed atoms in constraints and atoms in '
           'structure have different lengths')
    with pytest.raises(TextXError, match=msg):
        meta_model.model_from_str(inp, **model_kwargs)
